using DataStructures

"""
    in(node::Node, graph::DAG)

Check whether the node is part of the graph.
"""
in(node::Node, graph::DAG) = node in graph.nodes

"""
    in(edge::Edge, graph::DAG)

Check whether the edge is part of the graph.
"""
function in(edge::Edge, graph::DAG)
    n1 = edge.edge[1]
    n2 = edge.edge[2]
    if !(n1 in graph) || !(n2 in graph)
        return false
    end

    return n1 in children(n2)
end

"""
    is_dependent(graph::DAG, n1::Node, n2::Node)

Returns whether `n1` is dependent on `n2` in the given `graph`.
"""
function is_dependent(graph::DAG, n1::Node, n2::Node)
    if !(n1 in graph) || !(n2 in graph)
        return false
    end
    if n1 == n2
        return false
    end

    queue = Deque{Node}()
    push!(queue, n1)

    # TODO: this is probably not the best way to do this, deduplication in the queue would help
    while !isempty(queue)
        current = popfirst!(queue)
        if current == n2
            return true
        end

        for c in current.children
            push!(queue, c)
        end
    end

    return false
end

"""
    pairwise_independent(graph::DAG, nodes::Vector{Node})

Returns true if all given `nodes` are independent of each other, i.e., no path exists from any member of `nodes` to any other member.

See [`is_dependent`](@ref)
"""
function pairwise_independent(graph::DAG, nodes::Vector{<:Node})
    checked_set = Vector{Node}()
    for n in nodes
        for c in checked_set
            if is_dependent(graph, c, n) || is_dependent(graph, n, c)
                return false
            end
        end
        push!(checked_set, n)
    end
    return true
end
