"""
    Argument

Representation of one of the arguments of a [`AbstractComputeTask`](@ref) in a Node.
Has a member `id::UUID`, which is the id of the compute task's parent the argument refers to, and an `index`, which is the index in the symbol's object. If `index` is 0, the whole object is the argument.
A compute task can have multiple arguments.
"""
struct Argument
    id::Base.UUID
    index::Int64
end

"""
    AbstractTask

The shared base type for any task.
"""
abstract type AbstractTask end

"""
    AbstractComputeTask

Base type for all compute tasks.

!!! note 
    A subtype of this *must* be templated with its [`AbstractTaskFunction`](@ref)s!

See also: [`ComputeTask`](@ref), [`FusedComputeTask`](@ref), [`VectorizedComputeTask`](@ref)
"""
abstract type AbstractComputeTask end

"""
    AbstractTaskFunction

The base type for task functions used to dispatch computes.
"""
abstract type AbstractTaskFunction end

"""
    ComputeTask <: AbstractTask

A compute task. Has a member [`AbstractTaskFunction`](@ref), and a list of [`Argument`](@ref)s.
"""
struct ComputeTask{TaskFunction <: AbstractTaskFunction} <: AbstractComputeTask
    func::TaskFunction
    arguments::Vector{Argument}
end

"""
    AbstractDataTask <: AbstractTask

The shared base type for any data task.
"""
abstract type AbstractDataTask <: AbstractTask end

"""
    DataTask <: AbstractDataTask

Task representing a specific data transfer.
"""
struct DataTask <: AbstractDataTask
    data::Float64
end

"""
    FusedComputeTask <: AbstractComputeTask

A fused compute task made up of the computation of first `T1` and then `T2`.

Also see: [`get_types`](@ref).
"""
struct FusedComputeTask{TaskFunction1 <: AbstractTaskFunction, TaskFunction2 <: AbstractTaskFunction} <:
       AbstractComputeTask
    first_func::TaskFunction1
    second_func::TaskFunction2
    t1_output::Symbol
    first_func_arguments::Vector{Argument}
    second_func_arguments::Vector{Argument}
end

"""


"""
struct VectorizedComputeTask{TaskFunction <: AbstractTaskFunction} <: AbstractComputeTask
    task_type::TaskFunction
    arguments::Vector{Argument}
end
